"""
#################################
Miscellaneous (:mod:`verry.misc`)
#################################

This module provides some utilities, primarily for library developers.

.. autosummary::
    :toctree: generated/

    FormatSpec

"""

from .formatspec import FormatSpec

__all__ = ["FormatSpec"]
