def cadd(lhs: float, rhs: float) -> float: ...
def cmul(lhs: float, rhs: float) -> float: ...
def cdiv(lhs: float, rhs: float) -> float: ...
def csqr(value: float) -> float: ...
def fsqr(value: float) -> float: ...
